.nr VS 16
.vs 16p
.nr PS 12
.ps 12p
.EQ
delim $$
gsize 12
.EN
.LP
\&
.sp -1
.ll 6.5i
\l'6.5i'
.sp
.in 1.5i
\s18\fBA System for Pattern Matching
.br
Applications on Biosequences\fR\s0
.sp 1.5
\s13\fIGerhard Mehldau and Gene Myers*\fR\s0
.in -1.5i
\l'6.5i'
.FS
*Department of Computer Science,
University of Arizona,
Tucson, AZ  85721.
.FE
.nr VS 14
.vs 14p
.sp 1.5
\fBAbstract\fR
.ce 0
.LP
\fI
This paper presents the design and implementation of ANREP, a system
for finding matches to network expressions composed of ``spacers''
and approximate matches to network expressions.
A user specifies such patterns via a declarative, free-format, and
strongly typed language called A that
is presented here in a tutorial style through a series of progressively
more complex examples.
The sample patterns are for protein and DNA sequence motifs, the application
domain for which ANREP was specifically created.
ANREP provides a unified framework for almost all previously proposed
motif patterns and extends them by providing approximate matching,
a feature heretofore unavailable except for the limited case of
individual sequences.
The performance of ANREP is discussed and
an appendix gives a concise specification of syntax and semantics.
A portable C software package implementing ANREP
is available via anonymous remote file transfer.
\fR
.sp
\fBIntroduction\fR
.LP
In this paper we present a prototype system, $ANREP$, for defining and
searching for a class of patterns that should facilitate research into
motifs ranging from relatively simple promoter patterns (Hawley and McClure,
1983) or zinc finger patterns (Miller \fIet al.\fR, 1985) to more complex
motifs such as the one recently posed for cytosine methyltransferases
(Posfai \fIet al.\fR, 1989).
The system is capable of searching sequences for patterns that are a series
of approximate pattern matches separated by specifiable distance ranges.
The patterns are communicated to the system in a small language called $A$.
.PP
$ANREP$ extends previous work by providing a unified framework for concepts
found in isolation in earlier systems:
QUEST (Abarbanel \fIet al.\fR, 1984), a system for specifying character
patterns built around exacting matching to regular expressions,
Staden's software (1988), that allows for the logical combination of motifs
via a list of spatial constraints;
and the package by Gribskov \fIet al.\fR (1988), that computes and analyzes
profiles of positionally weighted protein motifs.
Furthermore, $ANREP$ is the first system capable of handling approximate
matches to network expressions rather than just keywords (as is possible in
some earlier systems).
.PP
At the top level, our system supports patterns that are network expressions
built from motif/threshold pairs and spacing constraints.
A network expression is a regular expression\p
.bp
without Kleene closure\u\s-1\(dg\s0\d,
.FS
\(dg Kleene closure is the repetition (zero or more times) of some substring
within a motif.
Kleene closure is commonly denoted by ``\fC*\fR'', e.g., the motif \&\fCac*a\fR
would match strings \&\fCaa\fR, \&\fCaca\fR, \&\fCacca\fR, \&\fCaccca\fR, ....
.FE
or equivalently, any expression built up of concatenation and union operators.
An example of such a pattern is given by the following specification:
.DS I 2
\f(CBnet\fC pattern = {A,1.0}\^(\^<0,20>\^{B,1.0}\^\||\|\^<-5,5>\^{C,1.0}\^)\^<25>\^{D,1.0};\fR
.DE
This network expression matches sequences consisting of some motif \&\fCA\fR,
followed by either motif \&\fCB\fR (starting somewhere between zero and
twenty symbols downstream) or motif \&\fCC\fR (five symbols upstream to
five symbols downstream), followed by motif \&\fCD\fR exactly twenty five
symbols downstream from the right end of the match for motif \&\fCB\fR
or \&\fCC\fR.  The 1.0 paired with each motif specifies the degree of
precision required to match it, a 1.0 asserting in this context
that the match must be exact.
.PP
Matches to motifs can be approximate as well as exact.
To define what we mean by an approximate match, first note that any
pattern is actually a specification of the strings it exactly matches.
A string approximately matches a pattern if it can be aligned, in the
sense of traditional sequence comparison, with one of the strings
specified by the pattern, in a manner that scores above a given
threshold under some scoring scheme for sequence alignments.
Note that an approximate match requires the specfication of a threshold
as well as a pattern.  Any string can be aligned to any of the strings
specified by the pattern, but not necessarily in a fashion whose
score is high enough to make it interesting or sufficiently close to
exact.  In $ANREP$, each motif is coupled with a threshold that defines
the stringency of the approximate match as seen in the example above.
Moreover, the scoring scheme used to evaluate approximate matches is
user-definable, and different scoring schemes may be associated with
different motifs or even subparts of motifs.
All approximate matches to a motif which exceed the associated threshold
are considered by the system in attempting to match the top-level
network expression of spacers and motif/threshold pairs.
.PP
In the language $A$, a motif
is a network expression built from ``atoms'' that
are either ASCII symbols or consensus symbols that generalize symbols
to weighted and/or tailored combinations of symbol sets.
An example is
.DS
\f(CBmotif\fC netexpr = "ac(gg\||\|ta)ct(gt)?";\fR
.DE
which specifies the four strings \&\fCacggct\fR, \&\fCacggctgt\fR,
\&\fCactact\fR, and \&\fCactactgt\fR.
The matching of motifs is approximate and so during the search process
each atom of a motif is conceptually
scored against the search string symbol (if any) with which it is aligned.
For symbols these scores are determined by a user-supplied sequence
comparison scoring scheme.
However, there is also the possibility of having an atom whose alignment
scores are tailored specifically for it, or are some admixture of the
schema scores for a set of symbols.
$ANREP$ provides such a capability that generalizes Staden's
weight matrices concept (Staden, 1988).
Finally, since matching is based on alignment scores one can also
value some portions of a match as being more important than others, i.e.,
one may positionally weight various parts of a motif.
\fIANREP\fP's positional weighting feature generalizes the capability
embodied in Gribskov's profile analysis software (Gribskov \fIet al.\fP,
1988).
.PP
The language $A$ is free-format, declarative, and strongly typed.
It permits patterns to be hierarchically defined and parameterized, supports
the developement of pattern libraries, and provides both enumerative and
functional mechanisms for specifying scoring schemes.
At the theoretical level, $ANREP$ is based on a new algorithm for finding
approximate matches to motifs, and a new optimized-backtracking method for
matching network expressions of motifs and spacers.
These algorithms are described in a forthcoming paper by the authors
(Mehldau and Myers, 1991).
The optimized backtracking approach is particularly robust.  In most
cases search time is relatively independent of pattern complexity,
as it is dominated by the time spent in an approximate search for
the least frequently occuring motif.
.PP
In the following sections we introduce the new system by giving a series of
examples.
The exact syntax and semantics for the language $A$ are presented in
Appendix A.
.sp
\fBSimple Patterns\fR
.LP
$ANREP$'s central function is to search a database file for instances
of user defined patterns, called \fInetworks\fP.  A network is a
network expression of \fImotifs\fP and \fIspacers\fP.  A motif, in turn,
is an approximate match to a network expression over a set of symbols.
Given a network and the name of a database file, $ANREP$ searches the
database for matches to the network and reports them in order of
appearance.
.PP
The most simple motif is a keyword or sequence of symbols.
An example of such an elementary pattern is the Pribnow box, "tata",
found in many promotor sequences.
To search for exact matches to this keyword, the following statement
suffices:
.DS
\f(CBsearch\fC DATABASE \f(CBfor\fC {"tata",1.0};\fR
.DE
The \&\f(CBsearch\fR statement initiates a scan of the contents of the
file \&\fCDATABASE\fR for all occurrences of the network
specified after the keyword \&\f(CBfor\fR.
In this instance the network expression is as simple as possible:
a motif/threshold pair.
The motif matches the literal string \&\fCtata\fR.
We defer the treatment of approximate matching to the next
section.  For now it is enough to know that a default scoring scheme is
automatically in effect and is such that a threshold of 1.0 requires that
the matches to the motif be exact.
Thus the example above searches \&\fCDATABASE\fR for all occurrences of
the string \&\fCtata\fR.
In general, the search statement operates on a file, specified directly
after the keyword \&\f(CBsearch\fR, and looks for matches to the
network specified after the keyword \&\f(CBfor\fR.
.PP
In general, an $ANREP$ network pattern is a network expression
of motifs and spacers.
The concatenation of two network expressions $P$ and $Q$, denoted
$PQ$, matches if there is a match to $P$,
directly followed by a match to $Q$.
The alternation of network expressions $P$ and $Q$, denoted $P^|^Q$,
matches if there is a match
to either $P$ or $Q$, and a match to a network expression $P$
can be made optional by following the $P$ with a question mark,
i.e., $P?$.
Alternation has the lowest precedence, concatenation the next lowest, and
optionality the highest.
Parentheses may be used for grouping, or to change the order of precedence.
A search for a pattern combining the so-called ``-35 sequence'' with
a Pribnow box 15 to 20 base pairs downstream, can be written in
$A$ as
.DS
\f(CBsearch\fC DATABASE \f(CBfor\fC {"ttgac",1.0} <15,20> {"tata",1.0};\fR
.DE
where the network consists of motif \&\fCttgac\fR,
concatenated with a spacer of the appropriate length,
concatenated with motif \&\fCtata\fR.
.PP
A spacer <$i$,$j$> matches at least $i$ and at most $j$ symbols.
If $j$ is omitted, the spacer matches exactly $i$ symbols.
Both $i$ and $j$ may be negative to indicate upstream rather than
downstream offsets.  For example,
.DS
\f(CBsearch\fC DATABASE \f(CBfor\fC {P,1.0} <-5,5> {Q,1.0};\fR
.DE
searches for an exact
match to motif \&\fCP\fR followed by an exact match to motif
\&\fCQ\fR, where the left end of the match to \&\fCQ\fR must start anywhere
from five symbols after, to five symbols before the right end
of the match to \&\fCP\fR.
.PP
Motifs are not restricted to keywords, but can be network expressions
of ``atomic'' symbols.  Regular ASCII symbols (like \&\fCa\fR, \&\fCc\fR,
\&\fCg\fR, or \&\fCt\fR) are the most simple and conventional atomic
symbols.
Examples of more complex atomic symbols (wild cards, classes,
and consensus symbols) are given in a later section.
.PP
Networks and motifs
(as well as alphabets, scoring schemes and numbers to be introduced) can be
assigned to variables, which may then be used in subsequent statements.
Variable names consist of any number of letters, digits, and underscores,
but must begin with a letter or an underscore.
For example, the sequence of statements:
.DS
\f(CBmotif\fC Pribnow = "tat?aa?t";\fR
\f(CBmotif\fC M35Seq  = "(tc)?ttgac(a\||\|t)";\fR
\f(CBnet\fC Promotor = {M35Seq,1.0} <15,20> {Pribnow,1.0};\fR
\f(CBsearch\fC DATABASE \f(CBfor\fC Promotor;\fR
.DE
assigns motifs to the variables \&\fCPribnow\fR and \&\fCM35Seq\fR;
then assigns a network involving these two motif variables to the variable
\&\fCPromotor\fR; and finally searches \&\fCDATABASE\fR for a match
to the network denoted by \&\fCPromotor\fP.
The motif \&\fCPribnow\fR matches one of the four strings \&\fCtaat\fR,
\&\fCtatat\fR, \&\fCtaaat\fR, and \&\fCtataat\fR, and the motif
\&\fCM35Seq\fR matches one of the strings \&\fCttgaca\fR, \&\fCttgact\fR,
\&\fCtcttgaca\fR, and \&\fCtcttgact\fR.
The network \&\fCPromotor\fP then matches an exact match to \&\fCM35Seq\fR
followed 15 to 20 symbols later by an exact match to \&\fCPribnow\fR.
.PP
Variables are defined hierarchically (i.e., they must be defined before
they are used), and may be parameterized.
For instance, using the definitions for \&\fCPribnow\fR and \&\fCM35Seq\fR
above, one can define and use a parameterized network $Ap_Promoter$
as follows:
.DS
\f(CBnet\fC Ap_Promoter{t1,t2} = {M35Seq,t1} <15,20> {Pribnow,t2};\fR
\f(CBnumber\fC thr = 0.75;\fR
\f(CBsearch\fC DATABASE \f(CBfor\fC Ap_Promoter{thr,thr};\fR
.DE
\&\fCAp_Promotor\fR is parameterized with the approximate match
thresholds for the two motifs in its network expression.
It is ``called'' in the subsequent \&\f(CBsearch\fR statement with the
same threshold for both motifs.
Under the scoring scheme that is in effect by default when ANREP is
first invoked, and given the threshold indicated, there can be one
symbol mismatched in a match to \&\fCPribnow\fR, and up to two
mismatches in an approximate match to \&\fCM35Seq\fR.
The exact nature of approximate matching will be explained in
the next section.
.sp
\fBAlphabets, Scoring Schemes, and Approximate Matches\fR
.LP
Alphabets are sets of symbols.
They are used to declare the range of symbols over which a scoring scheme,
and, by extension, a pattern is defined.
Each alphabet used in a pattern specification must cover the range of
symbols encountered in the database when searching for that pattern.
Examples are
.DS
\f(CBalpha\fC DNA = [acgt];\fR
\f(CBalpha\fC Protein = [ARNDCQEGHILKMFPSTWYV];\fR
.DE
which specify alphabets for DNA and protein sequences, respectively.
.PP
In the context of approximately matching a motif,
one must specify for each atomic motif symbol, what it costs to delete that
symbol, substitute some other symbol for it, and insert symbols in
the database sequence after it.
Traditionally these \fIedit costs\fP are specified by a single
global scoring scheme.
In most previous pattern matching systems, this underlying scoring scheme
was ``hard-wired'' into the program code.
However, for many applications it is important to be able to experiment with
different scoring schemes, either between different searches for the same
pattern, or between different sub-patterns over the length of a pattern.
For this reason, $ANREP$ provides for the explicit specification of scoring
schemes.
.PP
A scoring scheme consists of two components: an alphabet $SIGMA$, and a
$( | SIGMA |  + 1 )$ by $( |  SIGMA |  + 1 )$ scoring matrix $\(*d$ whose
indices range over the symbols in $SIGMA$ and the special character
$epsilon$ \o'\(sl\(mo' $SIGMA$.
The entries of $\(*d$ define the cost of the various editing operations
as follows.
For $x,^y~\(mo~SIGMA$, the cost of replacing $x$ with $y$ is $\(*d(^x,^y )$,
the cost of deleting $x$ is $\(*d(^x ,^epsilon )$, and the cost of
inserting $y$ is $\(*d(^epsilon ,^y )$.
The single entry $\(*d( epsilon ,^epsilon )$ is not used.
Figure 1 shows a scoring matrix for DNA sequences that scores insertions
and deletions (indels) $-1$, unequal substitutions $-.33$, and equal
substitutions $1$.
.TS
center delim($$);
r r r r r r.
	$epsilon$	\fCa\fR	\fCc\fR	\fCg\fR	\fCt\fR
$epsilon$		$-1$	$-1$	$-1$	$-1$
\fCa\fR	$-1$	1	$-.33$	$-.33$	$-.33$
\fCc\fR	$-1$	$-.33$	1	$-.33$	$-.33$
\fCg\fR	$-1$	$-.33$	$-.33$	1	$-.33$
\fCt\fR	$-1$	$-.33$	$-.33$	$-.33$	1
.TE
.ce
Figure 1:\ \ Sample scoring scheme
.sp
.PP
Consider the motif \fCtata(aaaa)?\fR which exactly matches either
\fCtata\fR or \fCtataaaaa\fR.  Call these strings the sequences of
the motif.  An approximate match to a motif is a substring that
aligns to one of the motif's sequences with a score that is
sufficiently large.  For example, in scanning a database we may
find the matches
.DS
\fC
 ... agttcagcatct-aataatccgacgacgatat-cgttactcgg ...
              tataaa-aa           tata
\fR
.DE
to the substrings \fCtctaataa\fR and \fCtat\fR.
The score of a match is the sum of the scores \(*d assigns to each
column of the alignment where insertions and deletions are introduced
by placing dashes in the appropriate sequence.  Thus, the score of
the first match is
$2\(*d(t,^t)^+^\(*d(a,^c)^+^\(*d(a,^\(*e)^+^4\(*d(a,^a)^+^\(*d(\(*e,^t)$
= $3.67$ and the score of the second is
$2\(*d(t,^t)^+^\(*d(a,^a)^+^\(*d(a,^\(*e)$ = 2.0.
A motif can actually match any substring of the database provided one is
not concerned with the score of the necessary alignment.  Thus one
is generally looking for matches whose score is above some threshold
of interest.  However, with motifs whose sequences vary in length,
giving an absolute threshold is problematic.  For example, if a threshold
of 5.0 were required to report matches to our sample motif, then only
matches to \fCtataaaaa\fR would ever be found since the best possible
match to \fCtata\fR scores 4.0.  That is, regardless of the scoring
scheme, matches to long sequences of the motif are always favored over
the shorter ones.
To rectify this, $ANREP$, compares a match's \fIlength relative score\fP
to the given threshold.  The length relative score is obtained by
dividing the score of the alignment by the length of the motif sequence
involved in it.
For example, the length relative scores of the matches
above are $.459$ and $.5$, respectively.  Further, the
motif/threshold pair, \fC{tata(aaaa)?,0.4}\fR reports matches
to \fCtataaaaa\fR whose absolute scores are above 3.2, reports matches
to \fCtata\fR scoring
above 1.6, and matches both of the substrings in the example above.
.PP
A \&\f(CBscore\fR statement can be used to name and define a scoring
scheme over a given alphabet.
Two formats are provided for defining the scoring matrix $\(*d$ \(em an
enumerative and a functional format.
Both formats may be combined within the same score declaration statement.
All entries not set explicitly have a value of $- inf$, i.e., not permitted.
The enumerative format requires giving a list of matrix entries
to be set, and the values they are to be set to.
This format is useful for scoring schemes like the Dayhoff metric (Dayhoff
1978), where values do not have an easily characterizable pattern.
An example of such a specification is the fragment
.DS
\f(CBscore\fC Dayhoff = Protein { <A.A>  #  0.90;
                          <A.R>  #  1.09;
                          <A.N>  #  0.99;
                          <A.D>  #  0.98;
                          <A.C>  #  1.12;
                           ...
                        };\fR
.DE
which declares a variable, \fCDayhoff\fR, of type scoring scheme over the
alphabet \fCProtein\fR, and assigns values to individual entries in the
semi-colon separated list between curly braces.  The entries to be set
are given between angle braces and the value to assign to them follows
a separating pound sign.
.PP
Within a pair of angle brackets, one can specify a set of matrix
entries in a number of ways.
The most basic specification is to give a comma-separated pair of characters,
\fC<x,y>\fR, which denotes the individual entry, $\(*d(^x,^y)$.
Both $x$ and $y$ must be from the underlying alphabet of the scoring
scheme, or they must be the ASCII symbol $"$"$ which stands for $epsilon$.
Either $x$ or $y$ (or both) can be replaced by a list of characters
between square brackets, in which case the denoted set of matrix
entries is the cross product of the sets of characters in each list,
e.g., \fC<[ac],[gt]>\fR) denotes the set of entries
\fC<a,g>\fR, \fC<a,t>\fR, \fC<c,g>\fR, and \fC<c,t>\fR.
If a period separates the pair instead of a comma, then the set
denoted is the ``symmetric'' cross product of the left and right
parts, e.g., \fC<A.R>\fR denotes the entries \fC<A,R>\fR and
\fC<R,A>\fR; and the notation \fC<[ag].t>\fR denotes the entries
\fC<a,t>\fR, \fC<g,t>\fR, \fC<t,a>\fR, and \fC<t,g>\fR.
Finally, the notation
\fC<x>\fR denotes the entries $<x,^y>$ where $y$ ranges over
all the symbols in $SIGMA$, i.e., it denotes the entire row of the matrix
labelled $x$ except for the entry $<x,^ epsilon >$.
An example making use of some of these forms is
.DS
\f(CBscore\fC Hybrid = DNA { <[acgt]>  #  -2;
                       <c.g>   #   3;
                       <a.t>   #   2;
                       <g.t>   #   1;
                   };\fR
.DE
that creates a ``hybridization'' matrix where
each pair of nucleotides that can stack is scored according to the
number of hydrogen bonds it contributes to the duplex, and non-stacking
pairs are scored $-2$.  Note that each specification should be thought
of as being ``executed'' in order of appearance.  Thus the entry for
$\(*d(a,^t)$ is first set to $-2$ in the first line, and later set
to 2 in the third line.  The matrix resulting from the above definition
is shown in Figure 2.
.TS
center delim($$);
r r r r r r.
	$epsilon$	\fCa\fR	\fCc\fR	\fCg\fR	\fCt\fR
$epsilon$		$- inf$	$- inf$	$- inf$	$- inf$
\fCa\fR	$- inf$	$-2$	$-2$	$-2$	2
\fCc\fR	$- inf$	$-2$	$-2$	3	$-2$
\fCg\fR	$- inf$	$-2$	3	$-2$	1
\fCt\fR	$- inf$	2	$-2$	1	$-2$
.TE
.ce
Figure 2:\ \ Hybridization matrix
.sp
.PP
For the functional format, the part to the left of the pound sign is
a boolean expression in two variables $x$ and $y$ that can take on
values from $SIGMA ^\(cu^"{" epsilon "}"$.  For every pair of
symbol choices for $x$ and $y$ for which this expression is true,
the entry $\(*d(x,^y)$ is set to the value given by the expression
in $x$ and $y$ to the right of the pound sign.
For example,
.EQ
delim @@
.EN
.DS
\f(CBscore\fC Sample = [acgt]
     { x!='$' && y!='$' # (4*(x==y) - 1) / 3.;
       < [acgt] . $ >   # -1.;
     };
.DE
which defines the sample scoring scheme for DNA sequences shown in
Figure 1.
All entries of the matrix \(*d, except for the first row and the first
column (where the expression \&\fCx != '$' && y != '$'\fR evaluates to false)
are set to the value of the expression to the right of the pound sign
which evaluates to 1.0 when \fIx\fP = \fIy\fP and \(mi.33 otherwise.
The \(*e entries are set by the second, enumerative clause to \(mi1.
.EQ
delim $$
.EN
.PP
In an $ANREP$ specification there is the concept of the
\fIcurrent scoring scheme\fP.  Initially, it is set by default
to the \fIunitary scoring scheme\fP, \&\fCUnitary\fR, over the alphabet of ASCII
characters.  The \(*d-matrix for this scheme does not permit indels,
scores identical substitutions 1, and mismatches 0.
Defining a scoring scheme has the effect of making it the current one.
It then applies to all motif definitions until another \&\f(CBscore\fR statement
is encountered.
The abbreviated statement
.DS
\f(CBscore\fC MyScheme;\fR
.DE
assumes \&\fCMyScheme\fR is a previously defined scoring scheme variable, and has
the effect of making it the current one.  This permits one to define a number
of schemes and have different ones be in effect for different portions of
the specification.
.PP
Recall that in the examples of the first section, it was asserted that
patterns of the form \&\fC{P,1.0}\fR required exact matches to
the network expression of the motif for \&\fCP\fR.  Since no scoring
schemes are defined in those examples, the default unitary scoring scheme
was in effect throughout.  Under this scheme the only way a match
can have a length relative score of 1.0 is for its absolute score
to be equal to the length of the motif word it matches, and the only
way this can happen is if the match is exact because only equal substitutions
score 1 or greater.  Thus in the default case, \&\fC{P,1.0}\fR denotes
an exact match to \&\fCP\fR.
This is also true for the \&\fCSample\fR scheme above, but not for
schemes like \&\fCHybrid\fR.
.PP
While the current scoring scheme is generally used to determine the
scoring of a motif declared within its scope, any part of a motif
may be scored under any other scheme by following the part in question
with an exclamation mark (``\&\fC!\fR'') and the name of the
scoring scheme to be applied to it.  The exclamation mark operator
has precedence between alternation and concatenation.
For example, for the motif \&\fCM35Seq\fR defined as follows:
.DS
\f(CBscore\fC Unitary;\fR
\f(CBmotif\fC M35Seq = "(tc)?(ttgac!Sample)((a\||\|t)t)?";\fR
.DE
an approximate match is scored with the default unitary scheme except
for the part of the match to the submotif \&\fCttgac\fR which is scored
with the sample scoring scheme of Figure 1.  For example, if the motif
sequence \&\fCtcttgacat\fR is matched to the sequence \&\fCactgccaat\fR as
follows:
.DS
\fC
     ac | t-gcca | at
     tc | ttgac- | at
\fR
.DE
then the part of the alignment between the bars is scored with \&\fCSample\fR
and the rest with \&\fCUnitary\fR.  The former scores $.67$ and the
latter two parts score $3$ for a length relative score of $.408$.
Note that the unaligned \&\fCa\fR is considered to be a part of the alignment
scored with \&\fCSample\fR.  In general, an unaligned character in the
database is scored according to the scheme associated with the first
motif symbol to its left.
.sp
\fBPositional Weights\fR
.LP
The concept of positional weights is introduced into $ANREP$
to let the user express the relative importance of different parts
of a motif.
With any submotif, one can associate a factor or weight by placing a
colon after the subexpression in question, followed by the weight,
an integer or real number in the range $\(+- 10.0 sup 10$.
The precedence of the colon operator is between that of an exclamation
mark and concatenation.
The score for aligning the given submotif is multiplied by the score
of the associated weight.  Thus weights of greater than 1 increase the
importance of a portion of a motif, and weights less than 1 decrease
their relative importance.
In addition, weight factors that are nested multiply.  That is, if
a subexpression weighted by \(*a is nested within another that is
weighted by \(*b, then the scores of the alignment for the inner motif
are multiplied by \(*a\(*b.
As an example of positional weighting consider,
.DS
\f(CBmotif\fC M35Seq = "(tc:.5)ttgac((a\||\|t:.5)t:.5)";\fR
.DE
which simply ``downgrades'' the less important parts of motif \&\fCM35Seq\fR as
opposed to making them entirely optional, as was done in the previous example.
The first sub-motif, \&\fCtc\fR, has a weight of 0.5; the second sub-motif,
\&\fCttgac\fR, has the default weight of 1.0; the third sub-motif,
\&\fCa\||\|t\fR, has a weight of 0.25 (nested weights), and the last sub-motif,
\&\fCt\fR, has a weight of 0.5.
.PP
When using positional weights, the relationship between scores, weights
and length-relative thresholds should be kept in mind.
Consider, for example,
.DS
\f(CBnet\fC Pribnow = {"ta(t:.5)a(a:.5)t",1.0};\fR
.DE
and
.DS
\f(CBnet\fC Pribnow = {"(ta:2)t(a:2)a(t:2)",1.0};\fR
.DE
which are not equivalent.
Assuming the default unitary scoring scheme and an input string (database)
of \&\fC...tatact...\fR, the first version matches with an absolute score
of 4.5, which translates into a length-relative score of 0.75.
In the second version, however, we get an absolute score of 9.0, and a
relative score of 1.50.
The threshold of 1.0 is therefore appropriate only for the second version.
To obtain identical results in the first version, its threshold should be
set to 0.5.
.sp
\fBAtoms\fR
.LP
Motifs are network expressions built up with the operators
\&\fC|\fR\|, \&\fC!\fR, \&\fC:\fR, justaposition, and \&\fC?\fR from
a collection of \fIatoms\fP which in all the examples thus
far have simply been ASCII characters.  The set of ASCII symbols
permitted in a motif are restricted to be those in the alphabet
of the scoring scheme associated with part of the motif the symbol
is in.  For instance, in the case of the \&\fCSample\fR scoring
scheme defined earlier, the legal symbols are those of the
alphabet \&\fCDNA\fR = $"{"a,^c,^g,^t"}"$.  In this section, we
specify and illustrate the ways $ANREP$ extends the notion of
atom in a number of powerful ways.
.sp
\fIThe Wild Card\fR
.LP
.PP
The wild card symbol, a period (``\&\fC.\fR'') in $ANREP$, is an atom
which is defined for all alphabets and matches any symbol from the
alphabet of the current scoring scheme.
An example is
.DS
\f(CBmotif\fC wild = "ca.t";\fR
.DE
which matches any one of the four strings \&\fCcaat\fR, \&\fCcact\fR,
\&\fCcagt\fR, and \&\fCcatt\fR (provided \&\fCDNA\fR is the alphabet of
the current scoring scheme).
Unlike all other atoms, the wild card does not contribute to the score
of a match, i.e., its substitution score is 0 for all letters.
However, its presence does contribute to the length of motif sequences
and thus it indirectly affects the length relative score of an
alignment involving it.
.sp
\fISymbol Classes\fR
.LP
A symbol class is a generalization from a single-symbol to a
multi-symbol atom.
A symbol class is specified as a set of symbols in square brackets,
and it matches any of the symbols in the set.
For example,
.DS
\f(CBmotif\fC class = "a[act]a";\fR
.DE
matches any one of the strings \&\fCaaa\fR, \&\fCaca\fR, and \&\fCata\fR.
Of course, this motif could also be written as
.DS
\f(CBmotif\fC class = "a(a\||\|c\||\|t)a";\fR
.DE
but not only is the symbol class easier to specify than the alternation
of symbols, it also is much more efficiently searched for.
.PP
It should be noted that the declaration
.DS
\f(CBmotif\fC wild = "ca[acgt]t";\fR
.DE
is \fInot\fR necessarily equivalent to the wild card example above,
because symbol classes as opposed to the wild card symbol do affect the
absolute score.  For example, if no substitution scores 0 in the current
scoring scheme, then the score of whatever the class matches will be
nonzero.
.sp
\fIConsensus Symbols\fR
.LP
In an effort to provide the user of the system with even finer control
over the motif to be matched, $ANREP$ further generalizes symbol
classes to consensus symbols.
Consensus symbols provide a way of specifying, at a given position in
a motif, exactly which symbols do or do not match, and their scores.
That is, they allow one to create a ``symbol'' whose edit costs are
either explicitly specified or are a function of the edit costs for
a set of characters.
.PP
In order to define the effect of consensus symbols and to show that they
generalize symbol classes, the concept of a \fIcost vector\fR is needed.
From the discussion of scoring schemes above, notice that out of the
entire scoring matrix, only two rows are relevant for any given symbol
$x$ in a motif.  Namely,
$\(*d(^epsilon ,^SIGMA )$, which gives the cost of inserting symbols
after $x$, and
$\(*d(^x,^SIGMA union roman "{" epsilon roman "}" )$, which gives 
the cost for deleting $x$ ($\(*d(^x,^epsilon )$), and substituting
another symbol for $x$ (the rest of the vector).
We call these two vectors the cost vectors associated with symbol $x$.
Figure 3 shows the cost vectors for symbol \&\fCg\fR, again assuming the
\&\fCSample\fR scoring scheme for DNA sequences.
.TS
center delim($$);
r r r r r r
| r r r r r r |
r r r r r r
r r r r r r
| r r r r r r |
r r r r r r.
	$epsilon$	\fCa\fR	\fCc\fR	\fCg\fR	\fCt\fR
_
$epsilon$		$-1$	$-1$	$-1$	$-1$
_
\fCa\fR	$-1$	1	$-.33$	$-.33$	$-.33$
\fCc\fR	$-1$	$-.33$	1	$-.33$	$-.33$
_
\fCg\fR	$-1$	$-.33$	$-.33$	1	$-.33$
_
\fCt\fR	$-1$	$-.33$	$-.33$	$-.33$	1
.TE
.ce
Figure 3:\ \ Cost vectors for symbol \&\fCg\fR
.sp
.PP
A consensus symbol is then defined as a compound symbol, whose
cost vectors are computed from the symbols involved in its definition.
Consensus symbols are specified in a manner similar to symbol
classes: a set of symbols and scores in square brackets.
There are currently four types of consensus symbols defined:
a ``\&\fC>\fR'' following the left bracket stands for a maximum reduction,
a ``\&\fC~\fR'' for average reduction, and a ``\&\fC<\fR'' for minimum reduction.
An ``\&\fC=\fR'' following the left bracket describes a tailored consensus
symbol.
.PP
The cost vectors of the
first three types of consensus symbols are obtained by applying
the appropriate reduction function to the cost vectors of the
set of symbols in its specification.
To make this clearer, look at the example:
.DS
\f(CBmotif\fC ConsensusSymbol = "[~at]";\fR
.DE
This motif simply consists of the consensus symbol \&\fC[~at]\fR, which is
to be computed as an average reduction of symbols \&\fCa\fR and \&\fCt\fR.
Figure 4 shows how the new cost vectors are computed from the old ones.
Minimum and maximum reductions are computed similarly, and it should be
clear that the principle could easily be extended to other reduction functions,
if they should found to be useful.
.TS
center delim($$);
r5 r r r r r7 r r r r r.
	$epsilon$	\fCa\fR	\fCc\fR	\fCg\fR	\fCt\fR	\
$epsilon$	\fCa\fR	\fCc\fR	\fCg\fR	\fCt\fR
\fCa\fR:		$-1$	$-1$	$-1$	\
$-1$	$-1$	1.0	$-.33$	$-.33$	$-.33$
\fCt\fR:		"	"	"	\
"	$-1$	$-.33$	$-.33$	$-.33$	1.0
						\
				
\fC[~at]\fR:		$- inf$	$- inf$	$- inf$	$- inf$	\
$- inf$	$.33$	$-.33$	$-.33$	$.33$
.TE
.ce
Figure 4:\ \ Computation of cost vectors for consensus symbol \&\fC[~at]\fR
.sp
.PP
When a consensus symbol is specified as in the previous example,
only substitution costs are included in the reduction \(em insertion
and deletion costs are set to $- inf$ irrespective of the original costs.
It is possible to include insertions and deletions in the computation,
as well as the rest of the symbols from the alphabet as a group.
Adding a \&\fC+\fR-sign to the class gives the consensus symbol an
insertion vector equal to the one for the symbols in the class (which
are all identical).  Adding a \&\fC-\fR-sign to the class, gives a deletion
cost that is the appropriate reduction of the deletion cost of the symbols
in the class (as opposed to $- inf$).  Finally, adding a \&\fC^\fR has the
effect of introducing every symbol not in the class, into the reduction
computation.  Alone this last addition is not of much value, but it is
also possible to weight the contribution of a symbol's vector by placing
a weight after the symbol separated with a colon.  As for positional
weights on motifs, these weights can be applied to sets of symbols and
may even nest if desired.
To illustrate the concepts, consider the example:
.DS
\f(CBmotif\fC Complex = "[> +:0.5 (ag:1.5) ^]!Sample";\fR
.DE
leads to the cost vectors shown in Figure 5.
Note that if a \&\fC-\fR sign was added after the \&\fC^\fR sign, then
the cost of deletion in the consensus would be $-1$ instead of
$- inf$.
.DS L
.ta .75iR 1.2iR 1.65iR 2.1iR 2.55iR 3.0iR 4.0iR 4.45iR 4.9iR 5.35iR 5.8iR
		\(*e	\fCa\fR	\fCc\fR	\fCg\fR	\fCt\fR	\(*e	\fCa\fR	\fCc\fR	\fCg\fR	\fCt\fR
.ta .75iR 1.2iR 1.65iR 2.1iR 2.55iR 3.0i+\w')*0.5'uR 4.0iR 4.45iR 4.9iR 5.35iR 5.8i+\w')*1.5'uR
	\fCa\fR:		($-1$	$-1$	$-1$	$-1$)*0.5	($-1$	1.0	$-.33$	$-.33$	$-.33$)*1.5
.ta .75iR 1.2iR 1.65iR 2.1iR 2.55iR 3.0iR 4.0iR 4.45iR 4.9iR 5.35iR 5.8i+\w')*1.5'uR
	\fCg\fR:		"	"	"	"	($-1$	$-.33$	$-.33$	1.0	$-.33$)*1.5
	\fCc\fR:		"	"	"	"	($-1$	$-.33$	1.0	$-.33$	$-.33$)*1.0
	\fCt\fR:		"	"	"	"	($-1$	$-.33$	$-.33$	$-.33$	1.0)*1.0
.sp .5
.ta .75iR 1.2iR 1.65iR 2.1iR 2.55iR 3.0iR 4.0iR 4.45iR 4.9iR 5.35iR 5.8iR
	\fCComplex\fR:		$-.5$	$-.5$	$-.5$	$-.5$	$- inf$	1.5	1.0	1.5	1.0
.DE
.ce
Figure 5:\ \ Computation of cost vectors for consensus symbol \&\fCComplex\fR
.sp
.PP
It should now be evident that consensus symbols are in fact a generalization
of symbol classes.
Specifically, \&\fC[act]\fR is equivalent to \&\fC[>act]\fR and is implemented
exactly this way.  This is significantly faster than \&\fC(a|g|c)\fR because
a single pair of cost vectors is built once for the consensus symbol, rather
than repeatedly having to take a three way minimum during the matching of
the network.  With the weighting feature and the special features
embodied in \&\fC+\fR, \&\fC-\fR, and \&\fC^\fR, $ANREP$ provides a rich
set of possible ways of summarizing the information content of a
position in a motif.
.PP
There is one final form of consensus symbol that permits one to
explicitly specify each element of the consensus' cost vectors.
An example of such a ``tailored'' consensus symbol is,
.DS
\f(CBmotif\fC Tailored = "[= +:-0.5 ag:1.5 ^:1.0]";\fR
.DE
In the case of a tailored consensus symbol, the weight associated with
a symbol is the score for matching that symbol.
For example, int the case of motif \&\fCTailored\fR, substituting an
\&\fCa\fR or \&\fCg\fR scores 1.5, substituting a \&\fCc\fR or \&\fCt\fR
scores 1.0; each symbol inserted immediately after it scorer $-0.5$,
and the symbol may not be deleted.
That is, the cost vectors for \&\fCTailored\fR are identical to that
of \&\fCComplex\fR above.
.PP
Tailored consensus elements permit one to specify arbitrary
weight matrices (Staden, 1988).  For example, Hawley and McClure (1983)
compiled the Pribnow sequence of 112 \fIE. Coli\fP promotors.
While the consensus is \&\fCtataat\fP, as used in the patterns of
the first section, very few promotors exactly match this string.
A better summary of the information present in the promotor region
is a record of the percentage of times each character appears in
a given position of the promotor.  The motif below captures this
information in the weight matrix of consensus symbols below.
For example, the second consensus symbols asserts that an \&\fCa\fR
appears in the second position 94% of the time.  The length relative
score of a match of \&\fCHandM\fR to the consensus \&\fCtataat\fR
is 4.22/6.0 = $.703$.
.DS
\f(CBmotif\fC HandM = "[a:.02 c:.09 g:.10 t:.79]
               [a:.94 c:.02 g:.01 t:.03]
               [a:.26 c:.14 g:.16 t:.44]
               [a:.59 c:.13 g:.15 t:.13]
               [a:.50 c:.20 g:.13 t:.17]
               [a:.01 c:.03 g:.00 t:.96]";\fR
.DE
While the matrix above illustrates the concept of the position of
a consensus pattern as being tendencies to be certain symbol, many
investigators advocate that these tendencies be the log (to the
base of the alphabet size) of the odds ratio of each symbol.
For example, the consensus symbol for the first position of the
pattern above would be \&\fC[a:-1.82 c:-.74 g:-.66 t:.83]\fR.  We will
not treat such information based measures in detail here, we simply wish
to point out that $ANREP$ provides the capability to model such
patterns.
.PP
Note that consensus symbols, while inducing quite intricate cost
vectors, are conceptually just ``characters'' in a motif.  Thus
their scores are still subject to multiplication by positional weighting
factors, and the final score of a match to the motif containing them
will be normalized with respect to the length before being compared
against the threshold.
of the match.
.sp
\fBA Comprehensive Example\fR
.LP
To illustrate how all the concepts described above work together,
we give a complete example of how to specify and search for a pattern.
Posfai \fIet al.\fR (1989) describe the overall structure of cytosine
methyltransferases (MTases), and list the actual amino acid sequences
of thirteen bacterial DNA MTases.
We use this paper as an example to show how a relatively complex pattern
can be described easily and concisely in $A$, and searched for using
$ANREP$.
.PP
We start out by defining the alphabet over which the pattern is defined,
and setting the scoring scheme to be used.
For brevity, we omit the definition of the \&\fCPAM250\fR scoring scheme,
which is very similar to the fragment given earlier for the \&\fCDayhoff\fR
metric.
Instead, we will assume that it has been previously defined and saved
in a library file named \&\fCPAM250\fR.
.DS
\f(CBalpha \fCProtein = [ARNDCQEGHILKMFPSTWYV];\fR
\f(CBinclude \fCPAM250;  # make PAM250 the current scoring scheme\fR
.DE
.PP
The overall structure of the cytosine methyltransferases consists of
ten distinct motifs, numbered from \&\fCI\fR to \&\fCX\fR, and separated
by spacers of variable length (all numbers in this example are directly
derived from the paper by Posfai \fIet al.\fR).
This pattern would be described in $A$ as the following network:
.DS
\f(CBnet \fCMTase { t01,t02,t03,t04,t05,t06,t07,t08,t09,t10 } =
            {   I,t01}  < 1, 29>
            {  II,t02}  < 5, 10>
            { III,t03}  < 6, 24>
            {  IV,t04}  <-3, 31>
            {   V,t05}  < 9>
            {  VI,t06}  <11, 32>
            { VII,t07}  < 3, 11>
            {VIII,t08}  <84,272>
            {  IX,t09}  < 5, 15>
            {   X,t10};\fR
.DE
Thresholds for the individual motifs are parameters to the network,
they must be provided when the network is used in a search (or within
a regular expression as part of a larger network).
This approach allows for flexibility during the search, as the network
can be ``called'' with varying thresholds, depending on the desired
specificity of the search.
.PP
To show how the individual motifs described by Posfai \fIet al.\fR translate
into $A$ motifs, we use motif \&\fCI\fR as an example \(em the process
is similar for the other motifs.
The simplest way to represent the motif is to mechanically translate
invariant positions into the corresponding symbol, positions with two
or three different amino acids into classes of symbols, and everything
else into the wild card symbol.
Leading and trailing wild cards are omitted.
.DS
\f(CBmotif \fCI = "[ILM] [SD] [LF] F [SAC] G . [GM] [GA]
           [FLI] .. [GAS] .... G" ;\fR
.DE
This form of the pattern is equivalent to the description given by
Posfai \fIet al.\fR.
However, it is possible to refine the pattern's definition (using
positional weights and consensus symbols), resulting in an even better
``fit'' between pattern and the actual matches as shown in Posfai
\fIet al.\fR.
The refinements are developed below.
.PP
The relative importance of the various positions in a motif depends on
the ``conservedness'' of the amino acids at those positions, and can be
expressed with positional weights.
Invariant positions are given a weight of three, positions where only
two different amino acids occur are assigned a weight of two, and
positions with three different amino acids are not weighted at all
(i. e., use the default weight of 1.0).
The wild card symbol does not contribute to the score, and therefore
does not need to be weighted.
Adjacent positions with the same weight are collected into groups.
.DS
\f(CBmotif \fCI = "[ILM] ([SD] [LF] :2) (F:3) [SAC] (G:3) .
           ([GM] [GA] :2) [FLI] .. [GAS] .... (G:3)" ;\fR
.DE
.PP
To account for the different frequencies of the various amino acids at
non-conserved positions, we replace the symbol classes with consensus
symbols, and assign weights according to those frequencies.
The resulting consensus symbols are normalized by use of positional
weights, such that the most frequent amino acid has an absolute
score of 1.0.
.DS
\f(CBmotif \fCI = "([> I:7 (LM:3)]:.143)
           (([> S:7 D:6]:.143) ([> L:12 F]:.083) :2)
           (F:3)
           ([> S:7 A:5 C:1]:.143)
           (G:3)
           .
           (([> G:12 M:1]:.083) ([> G:10 A:3]:.100) :2)
           ([> F:6 L:4 I]:.167)
           ..
           ([> G:7 A:5 S]:.143)
           ....
           (G:3)" ;\fR
.DE
.PP
To compute the maximum possible score, we need to add the maximum possible
score for each position, and divide their sum by the number of positions.
For motif \&\fCI\fR, that evaluates as
(1+2+2+3+1+3+0+2+2+1+0+0+1+0+0+0+0+3)/18 = 1.167.
A reasonable threshold for a ``tight'' search would therefore be around
1.0, and for a more ``loose'' search between 0.8 and 0.9.
A similar computation needs to be done for the other motifs.
To actually do the search, we only need to specify the database file to
be searched, and call the network \&\fCMTase\fR with the appropriate
thresholds:
.DS
\f(CBsearch \fCGenBank \f(CBfor \fCMTase{0.95,..,..,..,..,..,..,..,..,..};\fR
.DE
.PP
$ANREP$ reports matches in the following format.
The first line lists the relevant portion of the database, enclosed in
brackets of the form \&\fIi\fC>\fR and \&\fC<\fIj\fR, where $i$ and $j$
are the positions of the first and last database symbol shown, respectively.
The database display is followed by one line for each matched motif, where
the lines are sorted by frontiers from left to right.
Each motif is shown as \&\fIx\fC: ..^^..\fR, where $x$ represents the
length-relative score of the best match to that particular motif.
The circumflex symbols are positioned under those database symbols giving the
best match, and the period symbols indicate the total range of possible matches.
For example, the statement
.DS
\f(CBsearch\fC DATABASE \f(CBfor\fC {"ttgac",0.75} <15,20> {"tata",0.75};\fR
.DE
might result in output of the form
.DS
\fC
     81387> acttgacacctgactgcgtctgactatag <81415     
    1.0000: ..^^^^^..                    
    1.0000:                        .^^^^.
\fR
.DE
.sp
\fBDiscussion\fR
.LP
This paper introduced a new language, $A$, for specifying patterns ranging
from very simple to very complex, and a prototype system, $ANREP$, for
searching for approximate matches to those patterns.
In this section, we will discuss some of our experiences with $ANREP$.
.PP
The most obvious difference to previous systems is the expressive power and
versatility of the new system.
While $ANREP$ incorporates many features that are also found in older systems,
it adds many new concepts not found elsewhere.
Among those are the capability of searching for approximate matches to entire
networks, the ability to concisely specify alphabets and scoring schemes and
to attach them to different parts of a pattern, and last, but not least, the
flexibility provided for specifying atoms via consensus symbols and positional
weights.
.PP
In addition to the new language features, both the efficient approximate
match search algorithm, and the optimized-backtracking search strategy
have also proven to be very successful \(em the system turned out to be
quite fast.
Despite its ability to handle very complex patterns, $ANREP$ can search
more than 1 MB of data per minute (CPU time) for ``average'' patterns on
a VAX 8650 under 4.3 BSD UNIX, and even more on a modern workstation like
the SUN SparcStation 1.
Furthermore, the searching speed decreases only slightly as patterns become
more complex; it is, however, highly dependent on the specificity of a pattern.
.PP
While most of our experience with $ANREP$ points to a solid and useful system,
we also found some areas where further research is necessary or desirable:
An option to precompile patterns and to use the resulting pattern-specific
code for database searches (rather than interpreting a pattern using
general-purpose code) may result in drastic speed improvements, and is very
useful in cases where large databases need to be searched for the same pattern.
Search times could be reduced even further if the system could recognize
and take advantage of the (frequent) special case of exact matches to motifs.
.PP
Another area of possible improvements concerns better support for frequently
occurring types of motifs.
While $A$ allows nearly unlimited flexibility in the specification of
symbols, more of this flexibility is desirable at the motif level.
In particular, a Kleene closure operator would be useful for internal
repeats within motifs, especially if combined with a possibility for
quantification (i.e., the ability to specify a certain number or range
of numbers of repeats).
Other potential additions could include shorthand notation for the complement
(in the sense of Watson-Crick base pairs) of a motif, permutations of
motifs, or lists of motifs and separators (again combined with some method
for quantification).
.PP
The class of patterns that can currently be specified with $ANREP$
could be enhanced by allowing ``back references'' within motifs, or
by allowing some form of ``rules'' being added to a motif specification.
All of the possible enhancements outlined above are among the topics
currently being researched and being considered for inclusion in a future
version of $ANREP$.
.sp
\fBReferences\fR
.sp 0.5
.in 0.5i
.ti -0.5i
Abarbanel, R.M., Wieneke, P.R., Mansfield, E., Jaffe, D.A., and Brutlag, D.L. (1984),
``Rapid searches for complex patterns in biological molecules,''
\fINucleic Acids Research\fR, Vol. 12, No. 1 (January 1984), pp. 263-280.
.ti -0.5i 
Dayhoff, M.O. (1978),
\fIAtlas of Protein Sequence and Structure,\fR Vol. 5, Suppl. 3,
National Biomedical Research Foundation, Washington, DC.
.ti -0.5i
Gribskov, M., Homyak, M., Edenfield, J., and Eisenberg, D. (1988),
``Profile scanning for three-dimensional structural patterns in protein sequences,''
\fICABIOS\fR, Vol. 4, No. 1 (March 1988), pp. 61-66.
.ti -0.5i
Hawley, D.K., and McClure, W.R. (1983),
``Compilation and analysis of Escherichia coli promoter DNA sequences,''
\fINucleic Acids Research\fR, Vol. 11, No. 8 (March 1988), pp. 2237-2255.
.ti -0.5i
Miller, J., McLachlan, A.D., and Klug, A. (1985),
``Repetitive zinc-binding domains in the protein transcription factor IIIA from Xenopus oocytes,''
\fIEMBO Journal\fR, Vol. 4, No. 6 (June 1985), pp. 1609-1614.
.ti -0.5i
Myers, E.W., and Miller, W. (1989),
``Approximate matching of regular expressions,''
\fIBulletin of Mathematical Biology\fR, Vol. 51, No. 1 (January 1989), pp. 5-37.
.ti -0.5i
Posfai, J., Bhagwat, A.S., Posfai, G., and Roberts, R.J. (1989),
``Predictive motifs derived from cytosine methyltransferases,''
\fINucleic Acids Research\fR, Vol. 17, No. 7 (April 1989), pp. 2421-2435.
.ti -0.5i
Sankoff, D., and Kruskal, J.B. (1983),
\fITime Warps, String Edits, and Macromolecules: The Theory and Practice of Sequence Comparison\fR,
Addison-Wesley, Reading, MA, 1983.
.ti -0.5i
Staden, R. (1988),
``Methods to define and locate patterns of motifs in sequences,''
\fICABIOS\fR, Vol. 4, No. 1 (March 1988), pp. 53-60.
.ti -0.5i
Ukkonen, E. (1985),
``Finding approximate patterns in strings,''
\fIJournal of Algorithms\fR, Vol. 6 (March 1985), pp. 132-137.
.ti -0.5i
Wagner, R.A., and J.I. Seiferas (1978),
``Correcting counter-automaton-recognizable languages,''
\fISIAM Journal on Computing\fR, Vol. 7, No. 3 (August 1978), pp. 357-375.
.in
.sp
.nr VS 12
.vs 12p
.nr PS 9
.ps 9p
.EQ
gsize 9
.EN
\fBAppendix A: Language Specifications\fR
.LP
The $ANREP$ system descirbed in the paper proper
has been implemented in the form of a
free-format, declarative computer language, $A$, for specifying patterns and
initiating database searches.
The language is strongly typed and allows for hierarchical and parameterized
definition of variables of type number, alphabet, scoring scheme, motif, and
network.
$ANREP$ provides an interactive as well as a batch-type environment for $A$,
and it supports libraries of predefined subcomponents.
.PP
This appendix serves as the defining document for the language $A$;
it describes the syntax using a modified BNF grammar, and precisely
states the semantic meaning of each construct.
.PP
The following notation is used.
The syntactic element being defined is listed, followed by a ``\(->'' and
its definition.
Syntactic elements listed on one line need to appear in the same order
in the program; alternatives are separated by ``\||\|''.
\&\f(CBBold\fR type indicates keywords that appear literally in the program.
Terminals are printed in \&\fCCourier\fR typeface, and regular expressions
of terminals are shown as quoted strings.  The notation for these
regular expression is precisely that of the \fIegrep\fP UNIX\u\s-3TM\s0\d
tool.  Non-terminals are printed in \fIitalics\fR.
A ``+'' after a syntactic element indicates one or more occurrences of this
element, and a ``*'' means zero or more occurrences.
Syntactic elements enclosed in square brackets are optional, i.e. they can
appear zero or one times.
Numbers and identifiers in the grammar are followed by a type in curly
braces or in parentheses.
Curly braces have a declarative meaning, i.e. the type is being defined,
whereas parentheses have an assertive meaning, i.e. the type is required.
.PP
A specification (or program) in $A$ consists of a sequence of
statements.
White space (blanks, tabs, and newlines) is insignificant,
except that newlines serve to terminate comments.
.DS
.ta 0.75i 1.25i 2.75i
\fIspec\fR	\(->	\fIstmt\fR*
.DE
.DS
.ta 0.75i 1.25i 2.75i
\fIstmt\fR	\(->	\fIcomment\fR
.br
	\(or	\fInumdecl\fR
.br
	\(or	\fIalpdecl\fR
.br
	\(or	\fImotdecl\fR
.br
	\(or	\fInetdecl\fR
.br
	\(or	\fIscore\fR
.br
	\(or	\fIsearch\fR
.br
	\(or	\fIinclude\fR
.br
	\(or	\fIshow\fR
.br
	\(or	\fIexit\fR
.DE
.PP
Comments may be added anywhere to annotate specifications.
Comments start with the pound sign (``\fC#\fR''), and are terminated
by the end-of-line.
.DS
.ta 0.75i 1.25i 2.75i
\fIcomment\fR	\(->	\&\fC# \fR".*\en"
.DE
.sp
\fIVariables\fR
.LP
Variable names must begin with a letter or an underscore, and may be
followed by any number of letters, digits, and underscores.
Variable names are case-sensitive.
.DS
.ta 0.75i 1.25i 2.75i
\fIid\fR	\(->	"[\fCa\fR-\fCzA\fR-\fCZ_\fR][\fCa\fR-\fCzA\fR-\fCZ_0\fR-\fC9\fR]*"
.DE
.sp
\fINumbers\fR
.LP
Numbers are either integer or floating point quantities.
Integers are numbers without a decimal point or exponent.
As a special case, ``\fC+\fR'' and ``\fC-\fR'' are interpreted as shorthand
for +1 and -1, respectively.
Floating point numbers contain a decimal point or an exponent (or both).
Numbers can be assigned to variables via the number declaration statement.
.DS
.ta 0.75i 1.25i 2.75i
\fInumdecl\fR	\(->	\&\f(CBnumber\fR \fIid\fR{int} \&\fC=\fR \fInum\fR(int)\fC;\fR
.br
	\(or	\&\f(CBnumber\fR \fIid\fR{flp} \&\fC=\fR \fInum\fR(flp)\fC;\fR
.DE
.DS
.ta 0.75i 1.25i 2.75i
\fInum\fR{int}	\(->	\fIid\fR(int)
.br
	\(or	\fIicon\fR
.DE
.DS
.ta 0.75i 1.25i 2.75i
\fInum\fR{flp}	\(->	\fIid\fR(flp)
.br
	\(or	\fIfcon\fR
.DE
.DS
.ta 0.75i 1.25i 2.75i
\fIicon\fR	\(->	\fIint\fR
.br
	\(or	"[\fC+-\fR]"
.DE
.DS
.ta 0.75i 1.25i 2.75i
\fIfcon\fR	\(->	\fIint\fR "[\fCEe\fR]" \fIint\fR
.br
	\(or	\fIrat\fR [ "[\fCEe\fR]" \fIint\fR ]
.DE
.DS
.ta 0.75i 1.25i 2.75i
\fIint\fR	\(->	[ "[\fC+-\fR]" ] "[\fC0\fR-\fC9\fR]+"
.DE
.DS
.ta 0.75i 1.25i 2.75i
\fIrat\fR	\(->	[ "[\fC+-\fR]" ] "([\fC0\fR-\fC9\fR]+\fC.\fR[\fC0\fR-\fC9\fR]*\||\|\fC.\fR[\fC0\fR-\fC9\fR]+)"
.DE
.sp
\fIAlphabets\fR
.LP
Alphabets are sets of symbols, where symbols can be upper- or lower case
letters (represented by themselves), or any other ASCII character
(represented by a back slash followed by the character itself).
The alphabet declaration statement is used to assign alphabets to variables.
.DS
.ta 0.75i 1.25i 2.75i
\fIalpdecl\fR	\(->	\&\f(CBalpha\fR \fIid\fR{alp} \&\fC=\fR \fIalp\fC;\fR
.DE
.DS
.ta 0.75i 1.25i 2.75i
\fIalp\fR	\(->	\fIid\fR{alp}
.br
	\(or	\&\fC[\fIsymb\fR+\fC]\fR
.DE
.DS
.ta 0.75i 1.25i 2.75i
\fIsymb\fR	\(->	"[\fCa\fR-\fCzA\fR-\fCZ\fR]\||\|\fC\e\fR."
.DE
.sp
\fIAtoms\fR
.LP
An atom can be a consensus symbol, a regular symbol, or the special wild card
symbol ``\fC.\fR'' (which matches all symbols from the current alphabet).
.DS
.ta 0.75i 1.25i 2.75i
\fIatom\fR	\(->	\fIcons\fR
.br
	\(or	\fIsymb\fR
.br
	\(or	\&\fC.\fR
.DE
.sp
\fIConsensus Symbols\fR
.LP
A consensus symbol represents an artificial symbol, whose cost vectors
are computed from the literals involved in its definition.
Consensus symbols are specified in a syntactic form similar to alphabets:
.DS
.ta 0.75i 1.25i 2.75i
.br
\fIcons\fR	\(->	\&\fC[\fR [ "[\fC>~<=\fR]" ] \fIlit\fR \&\fC]\fR
.DE
.PP
The (optional) special symbol following the left square bracket indicates
how the cost vectors are to be computed from the literals and their weights.
A ``\fC>\fR'' stands for a maximum reduction, a ``\fC~\fR'' for an average
reduction, a ``\fC<\fR'' for a minimum reduction, and an ``\fC=\fR'' means
that the specification is to be taken literally \(em no transformation is
applied.
A missing special symbol denotes a symbol class and is equivalent to an
explicitly specified maximum reduction.
Following the special symbol, all literals involved in the definition of
the consensus symbol are listed \(em either by themselves, or as groups
enclosed in parentheses.
A literal can be either a regular symbol (which must be defined in the current
scoring scheme), or it can be one of the three special symbols
``\fC+\fR'', ``\fC-\fR'', and ``\fC^\fR'', which stand for insertion,
deletion, and substitution of all symbols not specified explicitly,
respectively.
Individual literals as well as groups of literals can be given a weight
by appending a colon and the desired weight.
Nested weights accumulate multiplicatively.
.DS
.ta 0.75i 1.25i 2.75i
\fIlit\fR	\(->	\fIlit \&\fC:\fR \fInum\fR(int\||\|flp)
.br
	\(or	\fIlit lit\fR
.br
	\(or	\&\fC(\fR \fIlit \&\fC)\fR
.br
	\(or	\fIsymb\fR
.br
	\(or	\&\fC+\fR
.br
	\(or	\&\fC-\fR
.br
	\(or	\&\fC^\fR
.DE
.PP
To describe the computation of the edit costs (cost vectors) for a
consensus symbol $alpha$, we define
$A^=~roman "{"^l sub i^:^w sub l sub i^ roman "}"$,
where $l$ is a literal from the specification, and $w sub l$ is its weight.
Furthermore, let $C$ = $A^-^"{" \fC+\fR, \fC-\fR, \fC^\fR "}"$,
i.e., the set of all regular symbols in the specification.
Substitution, deletion, and insertion costs are then defined for the case
of a reduction function $f$ as follows:
.TS
center delim($$);
l l.
.sp 0.125i
$\(*d(^alpha ,^c )$	$=~~~left {~~~matrix {\
lcol {{f (^f from down 16 {x \(mo C}^( \(*d(^x,^c ) w sub x ), \
f from down 16 {x \(mo SIGMA - C}^( \(*d(^x,^c ) w sub roman "^" ))}\
above {f from down 16 {x \(mo C}^( \(*d(^x,^c ) w sub x )}}\
lcol {{if~roman "^"^\(mo ^A} above {roman "otherwise"}}}$
.sp 0.25i
$\(*d(^alpha ,^epsilon )$	$=~~~left {~~~matrix {\
lcol {{f (^f from down 16 {x \(mo C}^( \(*d(^x,^epsilon ) w sub x ), \
f from down 16 {x \(mo SIGMA - C}^( \(*d(^x,^epsilon ) w sub - ))}\
above {f from down 16 {x \(mo C}^( \(*d(^x,^epsilon ) w sub x )}}\
lcol {{if~-^\(mo ^A} above {roman "otherwise"}}}$
.sp 0.25i
$\(*d(^epsilon ,^c)$	$=~~~left {~~~matrix {\
lcol {{\(*d(^epsilon ,^c ) w sub +} above {- inf}}\
lcol {{if~+^\(mo^A} above {roman "otherwise"}}}$
.TE
In the case of a tailored consensus symbol, these costs are defined as:
.TS
center delim($$);
l l.
.sp 0.125i
$\(*d(^alpha ,^c )$	$=~~~left {~~~matrix {\
lcol {{w sub c} above {w sub roman "^"} above {- inf}}\
lcol {{~~if^c^\(mo^C} above {~~if^c^\(mo back 70 \(sl \
fwd 40^C^and^roman "^"^\(mo^A} above {~~roman "otherwise"}}}$
.sp 0.25i
$\(*d(^alpha ,^epsilon )$	$=~~~left {~~~matrix {\
lcol {{w sub +} above {- inf}}\
lcol {{~~if^+^\(mo^A} above {~~roman "otherwise"}}}$
.sp 0.25i
$\(*d(^epsilon ,^c)$	$=~~~left {~~~matrix {\
lcol {{w sub -} above {- inf}}\
lcol {{~~if^-^\(mo^A} above {~~roman "otherwise"}}}$
.TE
.sp
\fIPatterns\fR
.LP
Patterns are regular expressions built from atoms, using the three operators
alternation (``\fC\||\|\fR''), concatenation, and optionality (``\fC?\fR'').
A particular scoring scheme can be applied to a specific pattern by appending
an exclamation mark and an identifier of type scoring scheme to the pattern.
To express the importance of a particular pattern relative to other patterns,
it may be followed by a colon and a weight.
Nested weights accumulate multiplicatively.
The order of precedence is ``\fC\||\|\fR'' (lowest), ``\fC!\fR'' and
``\fC:\fR'', concatenation, and ``\fC?\fR'' (highest).
Parentheses may be used for grouping and to force a different order of
precedence.
.DS
.ta 0.75i 1.25i 2.75i
\fIpat\fR	\(->	\fIpat \&\fC\||\|\fR \fIpat\fR
.br
	\(or	\fIpat \&\fC!\fR \fIid\fR(met)
.br
	\(or	\fIpat \&\fC:\fR \fInum\fR(int\||\|flp)
.br
	\(or	\fIpat pat\fR
.br
	\(or	\fIpat \&\fC?\fR
.br
	\(or	\&\fC(\fR \fIpat \&\fC)\fR
.br
	\(or	\fIatom\fR
.DE
.sp
\fIMotifs\fR
.LP
Motifs are regular expressions built from patterns and variables of type motif.
Motifs are assigned to variables via the motif declaration statement.
A motif declaration may be parameterized.
Parameters serve as placeholders for actual values that are filled
in when the motif is ``called'', or used.
.DS
.ta 0.75i 1.25i 2.75i
\fImotdecl\fR	\(->	\&\f(CBmotif\fR \fIid\fR{mot} [ \&\fC{\fR \fIid \fR( \&\fC,\fR \fIid \fR)* \&\fC}\fR ] \&\fC=\fR \fImot\fC;\fR
.DE
.PP
Motif regular expressions are built from atoms, using the three operators
alternation (``\fC\||\|\fR''), concatenation, and optionality (``\fC?\fR'').
A particular scoring scheme can be applied to a specific motif by appending
an exclamation mark and an identifier of type scoring scheme to the motif.
To express the importance of a particular motif relative to other motifs in
the same pattern, a motif may be followed by a colon and a weight.
Nested weights accumulate multiplicatively.
The order of precedence is ``\fC\||\|\fR'' (lowest), ``\fC!\fR'' and
``\fC:\fR'', concatenation, and ``\fC?\fR'' (highest).
Parentheses may be used for grouping and to force a different order of
precedence.
.DS
.ta 0.75i 1.25i 2.75i
\fImot\fR	\(->	\fImot \&\fC\||\|\fR \fImot\fR
.br
	\(or	\fImot \&\fC!\fR \fIid\fR(met)
.br
	\(or	\fImot \&\fC:\fR \fInum\fR(int\||\|flp)
.br
	\(or	\fImot mot\fR
.br
	\(or	\fImot \&\fC?\fR
.br
	\(or	\&\fC(\fR \fImot\fR \&\fC)\fR
.br
	\(or	\fIid\fR(mot) [ \&\fC{\fR \fIparam \fR( \&\fC,\fR \fIparam \fR)* \&\fC}\fR ]
.br
	\(or	\&\fC"\fR \fIpat \&\fC"\fR
.DE
.PP
Parameter values can be numbers, motifs, networks, and identifiers of
type scoring scheme.
.DS
.ta 0.75i 1.25i 2.75i
\fIparam\fR	\(->	\fInum\fR(int\||\|flp)
.br
	\(or	\fImot\fR
.br
	\(or	\fInet\fR
.br
	\(or	\fIid\fR(met)
.DE
.sp
\fINetworks\fR
.LP
A variable of type network is declared and assigned to with the network
declaration statement.
Like a motif declaration, a network declaration may be parameterized.
.DS
.ta 0.75i 1.25i 2.75i
\fInetdecl\fR	\(->	\&\f(CBnet\fR \fIid\fR{net} [ \&\fC{\fR \fIid \fR( \&\fC,\fR \fIid \fR)* \&\fC}\fR ] \&\fC=\fR \fInet\fC;\fR
.DE
.PP
Network regular expressions are built from motifs, spacers, and other networks,
using the three operators (in order of increasing precedence) alternation
(``\fC\||\|\fR''), concatenation, and optionality (``\fC?\fR'').
Parentheses may be used for grouping and to force a different order of
precedence.
.DS
.ta 0.75i 1.25i 2.75i
\fInet\fR	\(->	\fInet \&\fC\||\|\fR \fInet\fR
.br
	\(or	\fInet net\fR
.br
	\(or	\fInet \&\fC?\fR
.br
	\(or	\&\fC(\fR \fInet \&\fC)\fR
.br
	\(or	\fIterm\fR
.DE
.PP
A term can be a motif preceded by an opening curly brace, and followed by
a comma, a threshold, and a closing curly brace, or it can be a spacer or
a reference to a previously defined network (with actual values provided
for all parameters).
.DS
.ta 0.75i 1.25i 2.75i
\fIterm\fR	\(->	\&\fC{\fR \fImot \&\fC,\fR \fInum\fR(int\||\|flp) \&\fC}\fR
.br
	\(or	\&\fC<\fR \fInum\fR(int) [ \&\fC,\fR \fInum\fR(int) ] \&\fC>\fR
.br
	\(or	\fIid\fR(net) [ \&\fC{\fR \fIparam \fR( \&\fC,\fR \fIparam \fR)* \&\fC}\fR ]
.DE
.PP
A motif in a network specification must match with a score greater than or
equal to the specified threshold during the search of a database, in order
for the whole network to match.
A spacer $<^i,^j>:~^i,^j^>=^0$ between two networks $M$ and $N$ serves as a
separator:
the right end of network $M$ must be at least $min(^i,^j)$ and at most
$max(^i,^j)$ symbols to the right of the left end of network $N$.
Such a spacer is therefore equivalent to a network consisting of a single
motif of $i$ periods, followed by $j$ periods in parentheses, followed by
a question mark.
However, spacers may also contain negative numbers; these are interpreted
as distances to the left.
The notation $<^i>$ is accepted as shorthand for $<^i,^i>$.
Spacers at the beginning or the end of a network are ignored.
.sp
\fIScoring Schemes\fR
.LP
$ANREP$ contains the concept of a current scoring scheme, which applies to
all patterns that do not have a scoring scheme specified explicitly.
Initially, the current scoring scheme is the unitary scoring scheme over
the alphabet of ASCII characters.
Scoring schemes are defined and/or made current with the score definition
statement.
.DS
.ta 0.75i 1.25i 2.75i
\fIscore\fR	\(->	\&\f(CBscore\fR \fIid\fR{met} [ \&\fC=\fR \fIalp \&\fC{\fR \fIdecl\fR+ \&\fC}\fR ] \&\fC;\fR
.br
	\(or	\&\f(CBscore\fR \fIid\fR(met) \&\fC;\fR
.DE
.PP
With the first form of the score statement, an identifier of type scoring
scheme is declared, initialized according to the specification to the right
of the equal sign, and made the current scoring scheme.
The second form of the score statement (without the declaration part), simply
causes the current scoring scheme to be set to the one specified by the
identifier.
.EQ
delim @@
.EN
.DS
.ta 0.75i 1.25i 2.75i
\fIdecl\fR	\(->	\&\fC<\fR \fIset \fR[ "[\fC.,\fR]" \fIset \fR] \&\fC>	#\fR\ \ (\fIicon\fR\||\|\fIfcon\fR)\fC;\fR
.br
	\(or	\fIc_expr	\&\fC#\fR\ \ \fIc_expr\fC;\fR
.DE
.DS
.ta 0.75i 1.25i 2.75i
\fIset\fR	\(->	\&\fC$\fR
.br
	\(or	\fIsymb\fR
.br
	\(or	\&\fC[\fR (\fIsymb \fR\||\|\&\fC$\fR)+ \&\fC]\fR
.DE
.DS
.ta 0.75i 1.25i 2.75i
\fIc_expr\fR	\(->	"[^\&\fC#;\fR]*"
.DE
.EQ
delim $$
.EN
.PP
The first (or enumerative) format explicitly specifies which entries are
to be set, and their values.
A declaration of this form is interpreted as follows:
Let $set sub 1$ be the first set of symbols listed inside the angle brackets,
and $set sub 2$ the second set of symbols (the dollar sign as a special symbol
stands for $epsilon$).
If $set sub 2$ is not present in the declaration, it is assumed to be $SIGMA$.
Let $op$ be either ',' or '.' (assumed to be ',' if not present), and
$value$ the value of the numeric constant to the right of the pound sign.
Then, if $op^=^$',', for all $a$ \(mo $set sub 1$ and all $b$ \(mo
$set sub 2$, $\(*d(^a,^b)^=^value$.
If $op^=^$'.', $\(*d(^a,^b)^=^\(*d(^b,^a)^=^value$, for all $a$ \(mo
$set sub 1$ and all $b$ \(mo $set sub 2$.
.PP
The second (or functional) format uses Boolean conditions to specify the
entries to be set.
The C expression to the left of the pound sign is interpreted as a Boolean
expression in variables $x$ and $y$, where $x$ represents the row (first)
index and $y$ the column (second) index into matrix $\(*d$.
This Boolean expression is then applied to each element of the matrix $\(*d$.
Each entry for which the expression evaluates to true (i.e., is not equal to
zero), is set to the value of the second C expression (to the right of the
pound sign), which may also be a function of $x$ and $y$.
.sp
\fISearching\fR
.LP
A search for a network over a database is initiated with the search statement.
The file identifier is the name of the file containing the database.
.DS
.ta 0.75i 1.25i 2.75i
\fIsearch\fR	\(->	\&\f(CBsearch\fR \fIfile_id \&\f(CBfor\fR \fInet\fC;\fR
.DE
.DS
.ta 0.75i 1.25i 2.75i
\fIfile_id\fR	\(->	"[\fCa\fR-\fCzA\fR-\fCZ_./\fR]+"
.DE
.PP
The database is searched according to the algorithms described previously.
All matches to the network specification contained in the database are
reported, subject to two restrictions:
(1) The maximum length of a match is limited by the size of an internal buffer,
and (2), if the overlap between two adjacent matches exceeds some percentage of
the shorter matches' length, only the first match found is reported.
However, both the maximum length of a match as well as the overlap percentage
are compilation constants; they can easily be adjusted to suit any particular
need.
.sp
\fIOther Features\fR
.LP
To facilitate libraries of predefined components, the include statement is
provided.
When an include statement is encountered, the contents of the file specified
by the file identifier are read in and processed as if they were typed
on-line \(em with one exception: if an error is encountered, processing
skips everything following the offending statement until the end of the
included file.
Include statements may be nested.
.DS
.ta 0.75i 1.25i 2.75i
\fIinclude\fR	\(->	\&\f(CBinclude\fR \fIfile_id\fC;\fR
.DE
.PP
The show statement is provided as a help for ``debugging'' complex networks,
motifs, and scoring schemes.
Given an identifier of type number or scoring scheme, show simply lists its
value; identifiers of type alphabet, motif, or network are displayed as
trees in semi-graphical form.
.DS
.ta 0.75i 1.25i 2.75i
\fIshow\fR	\(->	\&\f(CBshow\fR \fIid\fC;\fR
.DE
.PP
Finally, the exit statement terminates an $A$ specification, and is
equivalent to an end-of-file.
Both forms of the exit statement are equivalent.
.DS
.ta 0.75i 1.25i 2.75i
\fIexit\fR	\(->	\&\f(CBexit\fC;\fR
	\(or	\&\f(CBquit\fC;\fR
.DE
